/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Yahoo! Mail Notifier code.
 *
 * The Initial Developer of the Original Code is
 * Michael Kraft.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// This is an object for settings and retrieving preference values.
// In addition it allows the assignment of observers and can set, retrieve or clear passwords.
// Portions of this code are based off of the preference code in the GreaseMonkey and 
// GMail Notifier Firefox extensions.

/*
 *  User info object
 */
function ym_user_info() {
  this.userName = null;
  this.password = null;
  this.remember_pw = false;
  this.yahoo_cookie = null;
  this.overrideURL = false
  this.mailURL = "http://mail.yahoo.com";
  this.useCookie = true;
}

function ym_prefs() {
  this.prefBranch = null;

  // pref values
  this.PREF_AUTO_CHECK            = "auto-check";
  this.PREF_CLICK                 = "click";
  this.PREF_DEFAULT_USER          = "users.default"; 
  this.PREF_DOUBLECLICK           = "doubleclick";
  this.PREF_DOUBLECLICK_TIME      = "doubleclicktime";
  this.PREF_LOG_ENABLED           = "logenable";
  this.PREF_LAST_LOGIN_TIME       = "lastlogintime";
  this.PREF_INTERVAL              = "interval";
  this.PREF_MAILURL               = "mailURL";
  this.PREF_MIDDLECLICK           = "middleclick";
  this.PREF_OVERRIDEURL           = "overrideURL";
  this.PREF_REMEMBER_PASSWORD     = "remember-password";
  this.PREF_SHOW_POPUP            = "show-popup";
  this.PREF_SHOW_ALERTS           = "show-alerts";
  this.PREF_SOUND                 = "sound";
  this.PREF_SOUNDFILE             = "soundfile";
  this.PREF_STATUSBAR             = "statusbar";
  this.PREF_YAHOOCOOKIE           = "yahoo-cookie"; 
  this.PREF_USERS                 = "users"; 
  this.PREF_VERSION               = "version";
  
  // current version
  this.CONST_VERSION     = "0.9.9.4";

  // alert constants
  this.CONST_NEVER       =  0;
  this.CONST_ONCE        =  1;
  this.CONST_REPEAT      =  2;

  // click constants
  this.CONST_ONECLICK    =  0;
  this.CONST_TWOCLICK    =  1;
  this.CONST_MIDCLICK    =  2;

  // tab/window constants
  this.CONST_NOTHING     =  0;
  this.CONST_SAMETAB     =  1;
  this.CONST_NEWTAB      =  2;
  this.CONST_FOCUSTAB    =  3;
  this.CONST_NEWWIN      =  4;
  this.CONST_CHECKMAIL   =  10;
  
  // statusbar constants
  this.CONST_ON_NEW_MAIL =  1;
  this.CONST_ALWAYS      =  2;
  
  // how often should extension relogin to Yahoo
  this.CONST_RELOGIN_TIME = 12;
  
  // password manager URL
  this.passURL = "chrome://ym_notifier";
}

/**
 * Returns a object to the extension's preference branch
 */
ym_prefs.prototype.getPrefBranch = function(){
  if (!this.prefBranch){ 
    this.prefBranch = Components.classes['@mozilla.org/preferences-service;1'].
                      getService(Components.interfaces.nsIPrefService).
                      getBranch("extensions.ym_notifier.");
  }

  return this.prefBranch;
}

/**
 * Delete user login information
 */
ym_prefs.prototype.deleteUser = function(userName) {
  var pref = this.getPrefBranch();
  var userBranch = this.PREF_USERS + "." + userName + ".";

  // remove user from users list
  var newUsers = new Array();
  var users = this.getPref(this.PREF_USERS).split(" ");
  for (var i=0; i < users.length; i++) {
    if (users[i] != userName) {
      newUsers.unshift(users[i]);
    }
  }
  this.setPref(this.PREF_USERS, newUsers.join(" "));

  if (userName == this.getPref(this.PREF_DEFAULT_USER)) {
    if (newUsers[0]) this.setPref(this.PREF_DEFAULT_USER, newUsers[0]);
    else pref.clearUserPref(this.PREF_DEFAULT_USER);
  }

  if (this.exists(userBranch + this.PREF_REMEMBER_PASSWORD)) {
    this.clearPassword(userName);
    this.deletePref(userBranch + this.PREF_REMEMBER_PASSWORD);
    this.deletePref(userBranch + this.PREF_YAHOOCOOKIE);
    this.deletePref(userBranch + this.PREF_LAST_LOGIN_TIME);
    this.deletePref(userBranch + this.PREF_OVERRIDEURL);
    this.deletePref(userBranch + this.PREF_MAILURL);
  }
}

/**
 * Save user login information
 */
ym_prefs.prototype.saveUser = function(userName, password, remember_password, overrideURL, mailURL, make_default) {
  var userBranch = this.PREF_USERS + "." + userName + ".";

  // initialize user preferences if not already set
  this.initPref(userBranch + this.PREF_YAHOOCOOKIE, "");
  this.initPref(userBranch + this.PREF_LAST_LOGIN_TIME, 0);

  if (remember_password && (password != "")) 
    this.setPassword(password, userName);
  else 
    this.clearPassword(userName);
  this.setPref(userBranch + this.PREF_REMEMBER_PASSWORD, remember_password);
  this.setPref(userBranch + this.PREF_OVERRIDEURL, overrideURL);
  this.setPref(userBranch + this.PREF_MAILURL, mailURL);

  if (make_default) this.setPref(this.PREF_DEFAULT_USER, userName);

  // add user to users list and position it based on make_default value
  var users = this.getPref(this.PREF_USERS)
  if (users) users = this.getPref(this.PREF_USERS).split(" ");
  else users = new Array();
  var foundUser = false;
  for (var i=0; i < users.length; i++) {
    if (users[i] == userName) {
      foundUser = true;
      break;
    }
  }

  if (!foundUser) {
    users.unshift(userName);
    users = users.sort();
    this.setPref(this.PREF_USERS, users.join(" "));
  }
}

/**
 * Save user cookie
 */
ym_prefs.prototype.saveUserCookie = function(cookie, userName) {
  // if no username specified retrieve default user info
  if (!userName || userName=="undefined") userName = this.getPref(this.PREF_DEFAULT_USER);

  if (userName) {
    var cookiePref = this.PREF_USERS + "." + userName + "." + this.PREF_YAHOOCOOKIE;
    var lastloginPref = this.PREF_USERS + "." + userName + "." + this.PREF_LAST_LOGIN_TIME;

    // get current date/time
    var now = new Date();

    this.setPref(cookiePref, cookie);
    if (cookie)
       // store last login in hours since 1970 (
       this.setPref(lastloginPref, parseInt(now.getTime()/3600000));
    else
       this.setPref(lastloginPref, 0);
  }
}

/**
 * Save user mailURL
 */
ym_prefs.prototype.saveUserMailURL = function(mailURL, userName) {
  // if no username specified retrieve default user info
  if (!userName || userName=="undefined") userName = this.getPref(this.PREF_DEFAULT_USER);
  if (userName) {
    var mailURLPref = this.PREF_USERS + "." + userName + "." + this.PREF_MAILURL;

    this.setPref(mailURLPref, mailURL);
  }
}

/**
 * Retrieve user login information
 */
ym_prefs.prototype.loadUser = function(userName, loadPw) {
  var userInfo = new ym_user_info();

  // if no username specified retrieve default user info
  if (!userName || userName=="undefined") userName = this.getPref(this.PREF_DEFAULT_USER);

  var userBranch = this.PREF_USERS + "." + userName + ".";

  // if username not stored (remember password pref missing) return null;
  try {
    // no username means no info
    if (userName) {
      userInfo.userName = userName;
      if (loadPw != false) userInfo.password = this.getPassword(userName);
      userInfo.remember_pw = this.getPref(userBranch + this.PREF_REMEMBER_PASSWORD);
    }
  }
  catch (ex) {
    userInfo = null;
  }

  if (userInfo) {
    
    try { 
      userInfo.overrideURL = this.getPref(userBranch + this.PREF_OVERRIDEURL);
    } catch (ex) { 
      this.initPref(userBranch + this.PREF_OVERRIDEURL, false);
    }

    try { 
      userInfo.mailURL = this.getPref(userBranch + this.PREF_MAILURL);
    } catch (ex) { 
      this.initPref(userBranch + this.PREF_MAILURL, "http://mail.yahoo.com");
    }

    try { 
      userInfo.cookie = this.getPref(userBranch + this.PREF_YAHOOCOOKIE);
      var last_login_time = this.getPref(userBranch + this.PREF_LAST_LOGIN_TIME);
      var now = new Date();

      // if over 12 hours since last login clear cookie to force normal login
      if ((parseInt(now.getTime()/3600000) - last_login_time) > this.CONST_RELOGIN_TIME)
        userInfo.useCookie = false;

    } catch (ex) { 
      this.initPref(userBranch + this.PREF_YAHOOCOOKIE, "");
      this.initPref(userBranch + this.PREF_LAST_LOGIN_TIME, 0);
    }
  }

  return userInfo;
}

/**
 * Whether a preference exists
 */
ym_prefs.prototype.exists = function(prefName) {
  var pref = this.getPrefBranch();
  return pref.getPrefType(prefName) != pref.PREF_INVALID;
}

/**
 * Returns the named preference, or defaultValue if it does not exist
 */
ym_prefs.prototype.getPref = function(prefName) {
  var pref = this.getPrefBranch();

  var prefType = pref.getPrefType(prefName);

  // underlying preferences object throws an exception if pref doesn't exist
  if (prefType == pref.PREF_INVALID) {
    throw new Error("Preference is not set: " + prefName);
  }
    
  switch (prefType) {
    case pref.PREF_STRING:
      return pref.getCharPref(prefName);
    case pref.PREF_BOOL:
      return pref.getBoolPref(prefName);
    case pref.PREF_INT:
      return pref.getIntPref(prefName);
  }

  return null;
}

/**
 * Sets the named preference to the specified value. values must be strings,
 * booleans, or integers.
 */
ym_prefs.prototype.setPref = function(prefName, value) {
  var prefType = typeof(value);
    
  switch (prefType) {
    case "string":
    case "boolean":
      break;
    case "number":
      if (value % 1 != 0) {
        throw new Error("Cannot set preference to non integral number");
      }
      break;
    default:
      throw new Error("Cannot set preference with datatype: " + prefType);
  }
    
  // underlying preferences object throws an exception if new pref has a 
  // different type than old one. i think we should not do this, so delete
  // old pref first if this is the case.
  //
  if (this.exists(prefName) && prefType != typeof(this.getPref(prefName))) {
    this.deletePref(prefName);
  }

  // set new value using correct method
  switch (prefType) {
    case "string":
      this.getPrefBranch().setCharPref(prefName, value);
      break;
    case "boolean":
      this.getPrefBranch().setBoolPref(prefName, value);
      break;
    case "number":
      this.getPrefBranch().setIntPref(prefName, Math.floor(value));
      break;
  }
}
  
/**
 * Deletes the named preference or subtree
 */
ym_prefs.prototype.deletePref = function(prefName) {
  this.getPrefBranch().deleteBranch(prefName);
}

/**
 * Initializes the default preference values. this should not be required since
 * the default values should be initiated by the defaults file, but keep this
 * here to be safe.
 */
ym_prefs.prototype.initPrefs = function() {
  var arePrefsUpToDate = this.exists(this.PREF_VERSION);
  if (!arePrefsUpToDate)
  {
    // set prefs if they don't exist
    this.initPref(this.PREF_AUTO_CHECK, true);
    this.initPref(this.PREF_CLICK, this.CONST_NOTHING);
    this.initPref(this.PREF_DEFAULT_USER, "");
    this.initPref(this.PREF_DOUBLECLICK, this.CONST_SAMETAB);
    this.initPref(this.PREF_DOUBLECLICK_TIME, 400);
    this.initPref(this.PREF_INTERVAL, 5);
    this.initPref(this.PREF_MIDDLECLICK, this.CONST_FOCUSTAB);
    this.initPref(this.PREF_SHOW_POPUP, this.CONST_NEVER);
    this.initPref(this.PREF_SHOW_ALERTS, this.CONST_NEVER);
    this.initPref(this.PREF_SOUND, this.CONST_NEVER);
    this.initPref(this.PREF_SOUNDFILE, "");
    this.initPref(this.PREF_STATUSBAR, this.CONST_ALWAYS);
    this.initPref(this.PREF_LOG_ENABLED, false);
    this.initPref(this.PREF_USERS, "");
  }

  // update preferences that should change with this version
  if (!arePrefsUpToDate || (this.getPref(this.PREF_VERSION) != this.CONST_VERSION))
  {
    this.setPref(this.PREF_VERSION, this.CONST_VERSION);
    this.fixPasswords();
  }
}

/**
 * If the preference doesn't already exist, set it
 */
ym_prefs.prototype.initPref = function(aPrefName, aDefaultValue) {
  var prefExists = this.exists(aPrefName);
  if (!prefExists)
    this.setPref(aPrefName, aDefaultValue);
}


/**
 * Add an observer to call the observer aObserver when preference aPrefName changes
 */
ym_prefs.prototype.addObserver = function(aPrefName, aObserver) {
  var myPrefs = this.getPrefBranch();
  var prefBranchInternal = myPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);

  if (prefBranchInternal)
    prefBranchInternal.addObserver(aPrefName, aObserver, false);
}

/**
 * Remove observer that calls the observer aObserver when preference aPrefName changes
 */
ym_prefs.prototype.removeObserver = function(aPrefName, aObserver) {
  var myPrefs = this.getPrefBranch();
  var prefBranchInternal = myPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);

  if (prefBranchInternal)
    prefBranchInternal.removeObserver(aPrefName, aObserver);
}

//
//  The following functions have nothing to do with the preference interface but
//  since they are preference related they are included here.
//

/**
 * Get the password associated with the given username.  If username not
 * passed, retrieve password associated with stored username.
 */
ym_prefs.prototype.getPassword = function(username) {
  // if no username specified retrieve default user info
  if (username == null) username = this.getPref(this.PREF_DEFAULT_USER);

  // no username means no password
  if (username == "") return '';

  // Get Password Manager (does not exist in Firefox 3.0)
  var CC_passwordManager = Components.classes["@mozilla.org/passwordmanager;1"]
     
  if (CC_passwordManager != null) {
	    
    // Used by all except Firefox 3.0
    var passwordManager = CC_passwordManager.getService(Components.interfaces.nsIPasswordManagerInternal);
      
    var host = {value:""};
    var user =  {value:""};
    var pass = {value:""}; 
   
    try {
      passwordManager.findPasswordEntry(this.passURL, username, "", host, user, pass);
      return (pass.value);
    }
    catch(ex) {
      dump(ex);
    }
  } 
  // Firefox 3.0 browser
  else {
    try {

      // Get Login Manager (Firefox 3.0)
      var passwordManager = Components.classes["@mozilla.org/login-manager;1"]
                            .getService(Components.interfaces.nsILoginManager);
	  
      // Find users for this extension 
      var logins = passwordManager.findLogins({}, this.passURL, "/", null);
      
      for (var i = 0; i < logins.length; i++) {
        if (logins[i].username == username) {
 	        return logins[i].password;
          break;
        }
      }
    }
    catch(ex) {
      dump(ex);
    }
  }
  return '';
}

/**
 * Store and associated a password with the given username.  If username not
 * passed, store and associate password with stored username.
 */
ym_prefs.prototype.setPassword = function(password, username) {
  // if no username specified retrieve default user info
  if (username == null) username = this.getPref(this.PREF_DEFAULT_USER);

  // set the password
  if (username != "") this.setClearPassword(username, password, true);
}

/**
 * Clear the password associated with the given username.  If username not
 * passed, clear password associated with stored username.
 */
ym_prefs.prototype.clearPassword = function(username) {
  // if no username specified retrieve default user info
  if (username == null) username = this.getPref(this.PREF_DEFAULT_USER);

  // clear the password
  this.setClearPassword(username, "", false);
}


/** 
 * Common code to set/clear password
 */
ym_prefs.prototype.setClearPassword = function(username, password, set) {
  // Get Password Manager (does not exist in Firefox 3.0)
  var CC_passwordManager = Components.classes["@mozilla.org/passwordmanager;1"]
     
  if (CC_passwordManager != null) {
    // Used by all except Firefox 3.0
    var passwordManager = CC_passwordManager.getService(Components.interfaces.nsIPasswordManager);
      
    // try to remove the current password, if it's there
    try {
      passwordManager.removeUser(this.passURL, username);
    } 
    catch (ex) {
      dump(ex);
    }
          
    if (set) passwordManager.addUser(this.passURL, username, password);
  }
  // Firefox 3.0 browser
  else {
    // Get Login Manager (Firefox 3.0)
    var passwordManager = Components.classes["@mozilla.org/login-manager;1"]
                          .getService(Components.interfaces.nsILoginManager);
                          
    var oldLogin = Components.classes["@mozilla.org/login-manager/loginInfo;1"]
                   .createInstance(Components.interfaces.nsILoginInfo);
                   
    var newLogin = Components.classes["@mozilla.org/login-manager/loginInfo;1"]
                   .createInstance(Components.interfaces.nsILoginInfo);

    oldLogin.init(this.passURL, "/", null, username, this.getPassword(username), "", "");
    newLogin.init(this.passURL, "/", null, username, password, "", "");
    
    // try to remove the current password, if it's there
    if (oldLogin.password != "") {
     passwordManager.removeLogin(oldLogin);
    } 
    
    if (set) passwordManager.addLogin(newLogin);
  }
}  

/* 
 * Password host changed in 0.9.4.4 so update it for stored passwords
 */
ym_prefs.prototype.fixPasswords = function() {
  var oldPassURL = "chrome://ym-notifier/";

  // Get Password Manager (does not exist in Firefox 3.0)
  var CC_passwordManager = Components.classes["@mozilla.org/passwordmanager;1"]
     
  if (CC_passwordManager != null) {
    // Used by all except Firefox 3.0
    var passwordManager = CC_passwordManager.getService(Components.interfaces.nsIPasswordManager);

    // ask the password manager for an enumerator:
    var e = passwordManager.enumerator;

    // step through each password in the password manager until we find the one we want:
    while (e.hasMoreElements()) {
      try {
        // get an nsIPassword object out of the password manager.
        // This contains the actual password...
        var pass = e.getNext().QueryInterface(Components.interfaces.nsIPassword);
        if (pass.host == oldPassURL) {
          // found it!
          passwordManager.addUser(this.passURL, pass.user, pass.password);
          passwordManager.removeUser(oldPassURL, pass.user);
        }
      } catch (ex) {}
    }
  }      
  // Firefox 3.0 browser
  else {
    // Get Login Manager (Firefox 3.0)
    var passwordManager = Components.classes["@mozilla.org/login-manager;1"]
                          .getService(Components.interfaces.nsILoginManager);

    var nsLoginInfo = new Components.Constructor("@mozilla.org/login-manager/loginInfo;1",
                      Components.interfaces.nsILoginInfo);

    // Find users for this extension 
    var logins = passwordManager.findLogins({}, oldPassURL, "/", null);
      
    for (var i = 0; i < logins.length; i++) {
      try {
        newLogin = new nsLoginInfo();
        newLogin.init(this.passURL, "/", null, logins[i].username, logins[i].password, "", "");
        passwordManager.modifyLogin(logins[i], newLogin);
      }
      catch (ex) {}
    }
  }  
} 